# Copyright 2018 RedHat Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import tempfile

from oslotest import base

from glance_store.common import utils


class TestUtils(base.BaseTestCase):
    def test_cooperative_reader_returns_bytes(self):
        with tempfile.TemporaryFile() as fd:
            reader = utils.CooperativeReader(fd)
            # Make sure CooperativeReader does not use cooperative_read instead
            # of its own read method.
            reader.read = utils.CooperativeReader.read
            out = reader.read(reader)
            self.assertEqual(out, b'')
